#include <postgres.h>
#include <fmgr.h>

#include <string.h>
#include <errno.h>
#include <ctype.h>

/* for krb5_principal2salt* ... */
#define KRB5_PRIVATE 1
#include <krb5.h>
/* from include/krb5/kdb.h */
/* Salt types */
#define KRB5_KDB_SALTTYPE_NORMAL        0
#define KRB5_KDB_SALTTYPE_V4            1
#define KRB5_KDB_SALTTYPE_NOREALM       2
#define KRB5_KDB_SALTTYPE_ONLYREALM     3
#define KRB5_KDB_SALTTYPE_SPECIAL       4
#define KRB5_KDB_SALTTYPE_AFS3          5

PG_FUNCTION_INFO_V1(enctype);
Datum enctype(PG_FUNCTION_ARGS)
{
	/* Args: char char *enctype */
	text *txtenc;
	char *enctype;
	krb5_context ctx;
	krb5_enctype etype;
	int ret, success = 0;
	
	txtenc = PG_GETARG_TEXT_P(0);
	enctype = palloc(VARSIZE(txtenc) - VARHDRSZ + 1);
	memset(enctype, 0, VARSIZE(txtenc) - VARHDRSZ + 1);
	memcpy(enctype, VARDATA(txtenc), VARSIZE(txtenc) - VARHDRSZ);
	
	ret = krb5_init_context(&ctx);
	if (ret != 0) {
		elog(ERROR, "enctype: krb5_init_context(): %s\n", error_message(errno));
		goto out0;
	}
	
	ret = krb5_string_to_enctype(enctype, &etype);
	if (ret != 0) {
		elog(ERROR, "enctype: krb5_string_to_enctype(): %s\n", error_message(ret));
		goto out1;
	}
	success = 1;
	
out1:
	krb5_free_context(ctx);
out0:
	pfree(enctype);
	if (success)
		PG_RETURN_INT32(etype);
	else
		PG_RETURN_NULL();
}

PG_FUNCTION_INFO_V1(salttype);
Datum salttype(PG_FUNCTION_ARGS)
{
	/* Args: char char *salttype */
	text *txtsalt;
	char *salttype;
	krb5_context ctx;
	krb5_int32 salt;
	int ret, success = 0;
	
	txtsalt = PG_GETARG_TEXT_P(0);
	salttype = palloc(VARSIZE(txtsalt) - VARHDRSZ + 1);
	memset(salttype, 0, VARSIZE(txtsalt) - VARHDRSZ + 1);
	memcpy(salttype, VARDATA(txtsalt), VARSIZE(txtsalt) - VARHDRSZ);
	
	ret = krb5_init_context(&ctx);
	if (ret != 0) {
		elog(ERROR, "salttype: krb5_init_context(): %s\n", error_message(errno));
		goto out0;
	}
	
	ret = krb5_string_to_salttype(salttype, &salt);
	if (ret != 0) {
		elog(ERROR, "salttype: krb5_string_to_salttype(): %s\n"
			"supported salttypes: normal, v4, norealm, onlyrealm, afs3",
			error_message(ret));
		goto out1;
	}
	success = 1;
	
out1:
	krb5_free_context(ctx);
out0:
	pfree(salttype);
	if (success)
		PG_RETURN_INT32(salt);
	else
		PG_RETURN_NULL();
}
	

PG_FUNCTION_INFO_V1(string2key);
Datum string2key(PG_FUNCTION_ARGS)
{
	/* Args: char *princ_str, char *pw, int enctype, int salttype */
	text *txtprinc, *txtpw;
	char *princ_str, *pw;
	text *pg_key;
	
	int ret, i;
	krb5_context ctx;
	krb5_int32 salt;
	krb5_enctype etype;
	krb5_keyblock key;
	krb5_principal princ;
	krb5_data pwdata, saltdata;
	const char *hex = "0123456789abcdef";
	
	txtprinc = PG_GETARG_TEXT_P(0);
	txtpw    = PG_GETARG_TEXT_P(1);
	etype    = PG_GETARG_INT32(2);
	salt     = PG_GETARG_INT32(3);
	princ_str = palloc(VARSIZE(txtprinc) - VARHDRSZ + 1);
	pw        = palloc(VARSIZE(txtpw   ) - VARHDRSZ + 1);
	memset(princ_str, 0, VARSIZE(txtprinc) - VARHDRSZ + 1);
	memset(pw       , 0, VARSIZE(txtpw   ) - VARHDRSZ + 1);
	memcpy(princ_str, VARDATA(txtprinc), VARSIZE(txtprinc) - VARHDRSZ);
	memcpy(pw       , VARDATA(txtpw   ), VARSIZE(txtpw   ) - VARHDRSZ);
	
	ret = krb5_init_context(&ctx);
	if (ret != 0) {
		elog(ERROR, "string2key: krb5_init_context(): %s\n", error_message(errno));
		goto out0;
	}
	
	ret = krb5_parse_name(ctx, princ_str, &princ);
	if (ret != 0) {
		elog(ERROR, "string2key: krb5_parse_name(): %s\n", error_message(ret));
		goto out1;
	}
	
	pwdata.length = strlen(pw);
	pwdata.data = pw;
	
	switch (salt) {
		case KRB5_KDB_SALTTYPE_NORMAL:
			ret = krb5_principal2salt(ctx, princ, &saltdata);
			if (ret) {
				elog(ERROR, "string2key: krb5_principal2salt(): %s\n", error_message(ret));
				goto out2;
			}
			break;
		case KRB5_KDB_SALTTYPE_V4:
			saltdata.length = 0;
			saltdata.data = 0;
			break;
		case KRB5_KDB_SALTTYPE_NOREALM:
			ret = krb5_principal2salt_norealm(ctx, princ, &saltdata);
			if (ret) {
				elog(ERROR, "string2key: krb5_principal2salt_norealm(): %s\n", error_message(ret));
				goto out2;
			}
			break;
		case KRB5_KDB_SALTTYPE_ONLYREALM:
		case KRB5_KDB_SALTTYPE_AFS3:
			saltdata.length = krb5_princ_realm(ctx, princ)->length;
			saltdata.data = malloc(saltdata.length + 1);
			if (saltdata.data == NULL) {
				elog(ERROR, "string2key: making salt: %s\n", error_message(ENOMEM));
				goto out2;
			}
			saltdata.data[saltdata.length] = '\0';
			memcpy(saltdata.data, krb5_princ_realm(ctx, princ)->data, saltdata.length);
			break;
		default:
			elog(ERROR, "string2key: making salt: %s\n", error_message(KRB5_KDB_BAD_SALTTYPE));
			goto out2;
			break;
	}
	
	if (salt == KRB5_KDB_SALTTYPE_AFS3)
		saltdata.length = SALT_TYPE_AFS_LENGTH;
	ret = krb5_c_string_to_key(ctx, etype, &pwdata, &saltdata, &key);
	if (salt == KRB5_KDB_SALTTYPE_AFS3)
		saltdata.length = krb5_princ_realm(ctx, princ)->length;
	if (ret != 0) {
		elog(ERROR, "string2key: krb5_string_to_key(): %s\n", error_message(ret));
		goto out3;
	}
	
	pg_key = palloc(VARHDRSZ + key.length * 2);
	VARATT_SIZEP(pg_key) = VARHDRSZ + key.length * 2;
	if (pg_key == NULL) {
		elog(ERROR, "string2key: generating output: %s\n", error_message(ENOMEM));
		goto out4;
	}
	
	for (i = 0; i < key.length*2; i++)
		VARDATA(pg_key)[i] = hex[(key.contents[i/2] >> ((i&1)?0:4)) & 0xF];
	
	/* finish */
out4:
	krb5_free_keyblock_contents(ctx, &key);
out3:
	krb5_free_data_contents(ctx, &saltdata);
out2:
	krb5_free_principal(ctx, princ);
out1:
	krb5_free_context(ctx);
out0:
	if (pg_key != NULL)
		PG_RETURN_TEXT_P(pg_key);
	else
		PG_RETURN_NULL();
}
