#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <com_err.h>

#include <libpq-fe.h>

#include <autoconf.h>
#include <krb5.h>

/* grr... why aren't these files distributed? */
#include <kadm5/admin.h>
#include <krb5/kdb.h>

typedef struct {
	krb5_context ctx;
	void *kadm;
	PGconn *conn;
	int verbose;
	const char *conninfo;
	const char *kerb_tbl, *pg_tbl;
	const char *username_col, *enctype_col, *salttype_col, *key_col;
} priv_data_t;

extern krb5_keyblock master_keyblock;

int hex2bin_c(unsigned int c)
{
	if ((c >= '0')&&(c <= '9'))
		return c - '0';
	if ((c >= 'A')&&(c <= 'F'))
		return c - 'A' + 10;
	if ((c >= 'a')&&(c <= 'f'))
		return c - 'a' + 10;
	return -1;
}

krb5_error_code hex2bin(const char *str, krb5_octet **bin, int *len)
{
	krb5_octet *p;
	int i, l;
	
	if (!bin)
		return EINVAL;
	
	for (i = 0; str[i] != '\0'; i++)
		if (hex2bin_c(str[i]) == -1)
			return EINVAL;
	
	l = i;
	if ((l & 1) != 0)
		return EINVAL;
	l /= 2;
	
	p = malloc(l);
	if (p == NULL)
		return ENOMEM;
	
	for (i = 0; i < l; i++)
		p[i] = hex2bin_c(str[i*2]) << 4 | hex2bin_c(str[i*2+1]);
	
	*bin = p;
	if (len)
		*len = l;
	
	return 0;
}

krb5_error_code bin2hex(krb5_octet *bin, int len, char **str)
{
	char *p;
	int i;
	
	if (!str)
		return EINVAL;
	
	p = malloc(len*2 + 1);
	if (p == NULL)
		return ENOMEM;
	
	for (i = 0; i < len; i++)
		snprintf(&(p[i*2]), 3, "%02x", bin[i]);
	
	*str = p;
	return 0;
}

void dopqclear(priv_data_t *pd, PGresult *res)
{
	PQclear(res);
}

PGresult *dopqexec(priv_data_t *pd, const char *query, int need_res)
{
	PGresult *res;
	
if (pd->verbose) printf("Executing query: %s\n", query);
	res = PQexec(pd->conn, query);
if (pd->verbose) printf("Result: %d, %s\n", PQstatus(pd->conn), PQerrorMessage(pd->conn));
	if (need_res)
		return res;
	
	dopqclear(pd, res);
	return NULL;
}

char *dopqescape(priv_data_t *pd, const char *val)
{
	size_t len;
	char *sqlval;
	
	len = strlen(val);
	sqlval = malloc (len * 2 + 1);
	PQescapeString(sqlval, val, len);
	
	return sqlval;
}

krb5_error_code init_krb(priv_data_t *pd)
{
	kadm5_config_params params;
	krb5_error_code ret;
	krb5_ccache cc;
	
printf("running init\n");
	ret = krb5_init_context(&(pd->ctx));
	if (ret != 0)
		goto out0;
	
printf("running cc_default\n");
	ret = krb5_cc_default(pd->ctx, &cc);
	if (ret != 0)
		goto out1;
	
printf("running kadm5_init\n");
	ret = kadm5_init_with_skey("(local)", NULL, KADM5_ADMIN_SERVICE, NULL,
		KADM5_STRUCT_VERSION, KADM5_API_VERSION_2, &(pd->kadm));
	if (ret != 0)
		goto out2;
	
printf("running get_config\n");
	ret = kadm5_get_config_params(pd->ctx, NULL, NULL, NULL, &params);
	if (ret != 0)
		goto out3;
	
printf("running db_set_name\n");
	ret = krb5_db_set_name(pd->ctx, params.dbname);
	if (ret != 0)
		goto out3;
	
printf("running db_init\n");
	ret = krb5_db_init(pd->ctx);
	if (ret != 0)
		goto out3;
	
printf("init_krb finished successfully\n");
	return 0;
	/* error handling */
out3:
	kadm5_destroy(pd->kadm);
out2:
	krb5_cc_close(pd->ctx, cc);
out1:
	krb5_free_context(pd->ctx);
out0:
	return ret;
}

kadm5_ret_t finish_krb(priv_data_t *pd)
{
	krb5_db_fini(pd->ctx);
	kadm5_destroy(pd->kadm);
	/*krb5_free_context(ctx);*/ /* SEGV here?? */
	return 0;
}

krb5_error_code dump_entry(krb5_pointer ptr, krb5_db_entry *ent)
{
	int i;
	char *q, *sqlname, *princname, *plainkeystr;
	krb5_error_code ret;
	krb5_keyblock plainkey;
	krb5_keysalt plainsalt;
	krb5_key_data *kd;
	krb5_data *name;
	priv_data_t *pd;
	
	pd = (priv_data_t *) ptr;
	
	if (krb5_princ_size(pd->ctx, ent->princ) != 1) {
		/* skip princs with instances (admins, services princs, and the like) */
		return 0;
	}
	name = krb5_princ_component(pd->ctx, ent->princ, 0);
	if ((name->length == 3) && !memcmp(name->data, "afs", 3)) {
		/* skip afs princ */
		return 0;
	}
	princname = malloc(name->length + 1);
	memcpy(princname, name->data, name->length);
	princname[name->length] = '\0';
	sqlname = dopqescape(pd, princname);
	
	for (i = 0; i < ent->n_key_data; i++) {
		kd = &(ent->key_data[i]);
		ret = krb5_dbekd_decrypt_key_data(pd->ctx, &master_keyblock, kd, &plainkey, &plainsalt);
		if (ret != 0) {
			printf("failed to decrypt key! skipping...\n");
			continue;
		}
		
		ret = bin2hex(plainkey.contents, plainkey.length, &plainkeystr);
		if (ret != 0) {
			printf("failed to encode key! skipping...\n");
			continue;
		}
		
		asprintf(&q, "INSERT INTO %s (%s, %s, %s, %s)"
			" VALUES ('%s', %d, %d, '%s');\n",
			pd->kerb_tbl,
			pd->username_col,
			pd->enctype_col,
			pd->salttype_col,
			pd->key_col,
			sqlname,
			kd->key_data_type[0],
			kd->key_data_type[1],
			plainkeystr);
		
		dopqexec(pd, q, 0);
		
		free(q);
		free(plainkeystr);
	}
	
	free(sqlname);
	free(princname);
	return 0;
}

krb5_error_code dump_all_entrys(priv_data_t *pd)
{
	krb5_error_code ret;
	int v;
	
	v = pd->verbose;
	pd->verbose = 0;
printf("Dumping all entries...\n");
	ret = krb5_db_iterate(pd->ctx, dump_entry, pd);
	pd->verbose = v;
	
	return ret;
}

krb5_error_code add_princ(priv_data_t *pd, char *name)
{
	krb5_error_code ret;
	kadm5_principal_ent_rec ent;
	krb5_db_entry dbent;
	krb5_boolean more;
	long mask;
	int cnt;
	
printf("Adding princ %s\n", name);
	/* Zero all fields in request structure */
	memset(&ent, 0, sizeof(ent));
	mask = 0;
	
	ret = krb5_parse_name(pd->ctx, name, &(ent.principal));
	if (ret != 0)
		goto out0;
	mask |= KADM5_PRINCIPAL;
	
	ent.attributes |= KRB5_KDB_DISALLOW_ALL_TIX;
	mask |= KADM5_ATTRIBUTES;
	
	ret = kadm5_create_principal(pd->kadm, &ent, mask, "foo");
	if (ret != 0)
		goto out1;
	
	ret = kadm5_randkey_principal(pd->kadm, ent.principal, NULL, NULL);
	if (ret != 0)
		goto out2;
	
	ent.attributes &= ~KRB5_KDB_DISALLOW_ALL_TIX; /* clear notix */
	mask = KADM5_ATTRIBUTES;
	
	ret = kadm5_modify_principal(pd->kadm, &ent, mask);
	if (ret != 0)
		goto out2;
	
	cnt = 1;
	ret = krb5_db_get_principal(pd->ctx, ent.principal, &dbent, &cnt, &more);
	if ((ret != 0)||(cnt != 1)||(more))
		goto out2;
	
	ret = dump_entry(pd, &dbent);
	if (ret != 0)
		goto out3;
	
	ret = 0; /* success */
out3:
	krb5_db_free_principal(pd->ctx, &dbent, cnt);
out2:
	if (ret != 0)
		kadm5_delete_principal(pd->kadm, ent.principal);
out1:
	krb5_free_principal(pd->ctx, ent.principal);
out0:
	return ret;
}

krb5_error_code del_princ(priv_data_t *pd, char *name)
{
	krb5_error_code ret;
	krb5_principal princ;
	char *q, *sqlname;
	
printf("Deleting princ %s\n", name);
	ret = krb5_parse_name(pd->ctx, name, &princ);
	if (ret != 0)
		goto out0;
	
	ret = kadm5_delete_principal(pd->kadm, princ);
	if (ret != 0)
		goto out1;
	
	sqlname = dopqescape(pd, name);
	asprintf(&q, "DELETE FROM %s WHERE %s = '%s'", pd->kerb_tbl, pd->username_col, sqlname);
	dopqexec(pd, q, 0);
	free(q);
	free(sqlname);
	
out1:
	krb5_free_principal(pd->ctx, princ);
out0:
	return ret;
}

kadm5_ret_t row2keysaltblock(priv_data_t *pd, PGresult *res, int row, krb5_keyblock *key, krb5_key_salt_tuple *ks)
{
	/* row format:  enctype salttype key */
	/* example row:  3 0 8910856eba8c */
	
	kadm5_ret_t ret;
	char *enctype, *salttype, *keystr, *rem;
	krb5_octet *keybin;
	krb5_enctype etype;
	krb5_int32 salt;
	int keylen;
	int enctype_col, salttype_col, key_col;
	
	enctype_col = PQfnumber(res, pd->enctype_col);
	salttype_col = PQfnumber(res, pd->salttype_col);
	key_col = PQfnumber(res, pd->key_col);
	
	if ((enctype_col == -1)||(salttype_col == -1)||(key_col == -1))
		return EINVAL;
	
	if (PQgetisnull(res, row, enctype_col)||
		PQgetisnull(res, row, salttype_col)||
		PQgetisnull(res, row, key_col))
		return EINVAL;
	
	enctype = PQgetvalue(res, row, enctype_col);
	salttype = PQgetvalue(res, row, salttype_col);
	keystr = PQgetvalue(res, row, key_col);
	
	if ((enctype[0] == '\0')||(salttype == '\0')||(keystr[0] == '\0'))
		return EINVAL;
	
	etype = strtol(enctype, &rem, 0);
	if (rem[0] != '\0')
		return EINVAL;
	
	salt = strtol(salttype, &rem, 0);
	if (rem[0] != '\0')
		return EINVAL;
	
	ret = hex2bin(keystr, &keybin, &keylen);
	if (ret != 0)
		return ret;
	
	key->enctype = etype;
	key->length = keylen;
	key->contents = keybin;
	ks->ks_enctype = etype;
	ks->ks_salttype = salt;
	
	return 0;
}

kadm5_ret_t do_setkey(priv_data_t *pd, char *name)
{
	/* row format:  enctype salttype key */
	/* example row:  3 0 8910856eba8c */
	/*              16 0 f132e38ccb0e4a91a4feb60831a20d3ea258c21a2613e580 */
	
	kadm5_ret_t ret;
	int i, count;
	krb5_principal princ;
	krb5_keyblock *keys;
	krb5_key_salt_tuple *ks;
	krb5_db_entry dbent;
	int cnt;
	krb5_boolean more;
	char *q, *sqlname;
	PGresult *res;
	
printf("Changing key for %s\n", name);
	sqlname = dopqescape(pd, name);
	asprintf(&q, "SELECT %s, %s, %s FROM %s WHERE %s = '%s'",
		pd->enctype_col, pd->salttype_col, pd->key_col,
		pd->pg_tbl, pd->username_col, sqlname);
	res = dopqexec(pd, q, 1);
	free(q);
	count = PQntuples(res);
	
	keys = malloc(sizeof(krb5_keyblock) * count);
	if (keys == NULL) {
		ret = ENOMEM;
		goto out0;
	}
	memset(keys, 0, sizeof(krb5_keyblock) * count);
	
	ks = malloc(sizeof(krb5_key_salt_tuple) * count);
	if (ks == NULL) {
		ret = ENOMEM;
		goto out1;
	}
	memset(ks, 0, sizeof(krb5_key_salt_tuple) * count);
	
	ret = krb5_parse_name(pd->ctx, name, &princ);
	if (ret != 0)
		goto out2;
	
	for (i = 0; i < count; i++) {
		ret = row2keysaltblock(pd, res, i, &(keys[i]), &(ks[i]));
		if (ret != 0)
			goto out3;
	}
	
	ret = kadm5_setkey_principal_3(pd->kadm, princ, 0, count, ks, keys, count);
	if (ret != 0)
		goto out3;
	
	asprintf(&q, "DELETE FROM %s WHERE %s = '%s'", pd->kerb_tbl, pd->username_col, sqlname);
	dopqexec(pd, q, 0);
	free(q);
	
	cnt = 1;
	ret = krb5_db_get_principal(pd->ctx, princ, &dbent, &cnt, &more);
	if ((ret != 0)||(cnt != 1)||(more))
		goto out3;
	
	ret = dump_entry(pd, &dbent);
	if (ret != 0)
		goto out4;
	
	ret = 0; /* success */
	
out4:
	krb5_db_free_principal(pd->ctx, &dbent, cnt);
out3:
	for (i = 0; i < count; i++)
		krb5_free_keyblock_contents(pd->ctx, &(keys[i]));
	krb5_free_principal(pd->ctx, princ);
out2:
	free(ks);
out1:
	free(keys);
out0:
	dopqclear(pd, res);
	free(sqlname);
	return ret;
}

void do_sync_del(priv_data_t *pd)
{
	PGresult *res;
	char *q;
	int i, cnt;
	int ret;
	
	asprintf(&q, "SELECT DISTINCT %s "
		"FROM %s p RIGHT JOIN %s k USING (%s) "
		"WHERE p.%s IS NULL;",
		pd->username_col,
		pd->pg_tbl, pd->kerb_tbl, pd->username_col,
		pd->username_col);
	res = dopqexec(pd, q, 1);
	free(q);
	
	cnt = PQntuples(res);
	for (i = 0; i < cnt; i++) {
		ret = del_princ(pd, PQgetvalue(res, i, 0));
		if (ret != 0)
			com_err("del_princ", ret, "deleting princ %s", PQgetvalue(res, i, 0));
	}
	
	dopqclear(pd, res);
}

void do_sync_add(priv_data_t *pd)
{
	PGresult *res;
	char *q;
	int i, cnt;
	int ret;
	
	asprintf(&q, "SELECT DISTINCT %s "
		"FROM %s p LEFT JOIN %s k USING (%s) "
		"WHERE k.%s IS NULL;",
		pd->username_col,
		pd->pg_tbl, pd->kerb_tbl, pd->username_col,
		pd->username_col);
	res = dopqexec(pd, q, 1);
	free(q);
	
	cnt = PQntuples(res);
	for (i = 0; i < cnt; i++) {
		ret = add_princ(pd, PQgetvalue(res, i, 0));
		if (ret != 0)
			com_err("add_princ", ret, "adding princ %s", PQgetvalue(res, i, 0));
			
	}
	
	dopqclear(pd, res);
}

void do_sync_key(priv_data_t *pd)
{
	PGresult *res;
	char *q;
	int i, cnt;
	int ret;
	
	/* userlist must be in sync befor this */
	asprintf(&q, "SELECT DISTINCT %s "
		"FROM %s p NATURAL FULL JOIN %s k "
		"WHERE p.%s IS NULL OR k.%s IS NULL;",
		pd->username_col,
		pd->pg_tbl, pd->kerb_tbl,
		pd->username_col, pd->username_col);
	res = dopqexec(pd, q, 1);
	free(q);
	
	cnt = PQntuples(res);
	for (i = 0; i < cnt; i++) {
		ret = do_setkey(pd, PQgetvalue(res, i, 0));
		if (ret != 0)
			com_err("do_setkey", ret, "setting key for princ %s", PQgetvalue(res, i, 0));
	}
	
	dopqclear(pd, res);
}

void do_sync(priv_data_t *pd)
{
printf("Syncing everything\n");
	do_sync_del(pd);
	do_sync_add(pd);
	do_sync_key(pd);
}

krb5_error_code init_conn(priv_data_t *pd)
{
	char *q;
	
	pd->conn = PQconnectdb(pd->conninfo);
	if (PQstatus(pd->conn) != CONNECTION_OK)
		return EINVAL;
	
	init_krb(pd);
	
	asprintf(&q, "CREATE TEMP TABLE %s ("
		"	%s varchar(8) NOT NULL,"
		"	%s int NOT NULL,"
		"	%s int NOT NULL,"
		"	%s text NOT NULL,"
		"	PRIMARY KEY (%s, %s, %s)"
		")",
		pd->kerb_tbl,
		pd->username_col,
		pd->enctype_col,
		pd->salttype_col,
		pd->key_col,
		pd->username_col, pd->enctype_col, pd->salttype_col
	);
	dopqexec(pd, q, 0);
	free(q);
	
	asprintf(&q, "LISTEN %s", pd->pg_tbl);
	dopqexec(pd, q, 0);
	free(q);
	
	return dump_all_entrys(pd);
}

krb5_error_code finish_conn(priv_data_t *pd)
{
	finish_krb(pd);
	PQfinish(pd->conn);
	
	return 0;
}

int main(int argc, char **argv)
{
	int ret;
	priv_data_t pd;
	PGnotify *note;
	fd_set rset;
	
	if (argc != 1) {
		fprintf(stderr, "Usage: %s\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	
	pd.verbose = 1;
	pd.conninfo = "dbname=hsguser user=sync_kerb";
	pd.kerb_tbl = "synckerb";
	pd.pg_tbl = "kerberos_keys";
	pd.username_col = "username";
	pd.enctype_col = "enctype";
	pd.salttype_col = "salttype";
	pd.key_col = "key";
	
	while (1) {
		init_conn(&pd);
		do_sync(&pd);
		while (1) {
			FD_ZERO(&rset);
			FD_SET(PQsocket(pd.conn), &rset);
			printf("waiting for activity on socket %d...\n", PQsocket(pd.conn));
			select(FD_SETSIZE, &rset, NULL, NULL, NULL);
			ret = PQconsumeInput(pd.conn);
			printf("got something\n");
			if (ret != 1) {
				printf("problem in consume input: %s\n", PQerrorMessage(pd.conn));
				printf("resetting....\n");
				break;
			}
			note = PQnotifies(pd.conn);
			if (note != NULL) {
				printf("got notification %s ==> syncing\n", note->relname);
				do_sync(&pd);
				free(note);
			}
		}
		finish_conn(&pd);
	}
	
	exit(EXIT_SUCCESS);
}

